#include <nemo/stdint.h>
#include <nemo/unixtypes.h>

#include <nemo/error.h>
#include <nemo/open.h>
#include <nemo/strerr.h>
#include <nemo/exit.h>
#include <nemo/stralloc.h>
#include <nemo/djbio.h>
#include <nemo/djbiofd.h>
#include <nemo/unix.h>

#include "cdb_make.h"
#include "cdb.h"

static const char FATAL[] = "cdbmake: fatal: ";

static char *fn;
static char *fntmp;

static void die_usage(void)
{
  strerr_die1x(100, "cdbmake: usage: cdbmake f ftmp");
}
static void die_nomem(void)
{
  strerr_die2x(111, FATAL, "out of memory");
}
static void die_create(void)
{
  strerr_die3sys(111, FATAL, "unable to create ", fntmp);
}
static void die_write(void)
{
  strerr_die3sys(111, FATAL, "unable to write ", fntmp);
}
static void die_rename(void)
{
  strerr_die5sys(111, FATAL, "unable to rename ", fntmp, " to ", fn);
}
static void die_read(void)
{
  strerr_die2sys(111, FATAL, "unable to read input");
}
static void die_readformat(void)
{
  strerr_die2x(111, FATAL, "unable to read input: bad format");
}

static void get(byte_t *ch)
{
  switch (djbio_GETC(djbiofd_in, ch)) {
    case 0: die_readformat();
    case -1: die_read();
    default: break;
  }
}

static struct cdb_make c;
static stralloc buf = STRALLOC;

int main(int argc, char **argv)
{
  unsigned int klen;
  unsigned int dlen;
  unsigned int i;
  uint32_t h;
  int fd;
  byte_t ch;

  if (argc < 3) die_usage();

  if (!*argv) die_usage();

  if (!*++argv) die_usage();
  fn = *argv;

  if (!*++argv) die_usage();
  fntmp = *argv;

  fd = open_trunc(fntmp);
  if (fd < 0) die_create();

  if (cdb_make_start(&c, fd) < 0) die_write();

  for (;;) {
    get(&ch);
    if (ch == '\n') break;
    if (ch != '+') die_readformat();
    klen = 0;
    for (;;) {
      get(&ch);
      if (ch == ',') break;
      if ((ch < '0') || (ch > '9')) die_readformat();
      if (klen > 429496720) {
        errno = error_nomem;
        die_write();
      }
      klen = klen * 10 + (unsigned int)(ch - '0');
    }
    dlen = 0;
    for (;;) {
      get(&ch);
      if (ch == ':') break;
      if ((ch < '0') || (ch > '9')) die_readformat();
      if (dlen > 429496720) {
	errno = error_nomem;
	die_write();
      }
      dlen = dlen * 10 + (unsigned int)(ch - '0');
    }

    if (cdb_make_addbegin(&c, klen, dlen) < 0) die_write();
    if (!stralloc_ready(&buf, klen)) die_nomem();
    if (!stralloc_erase(&buf)) die_nomem();
    for (i = 0; i < klen; ++i) {
      get(&ch);
      if (djbio_PUTC(&c.io, ch) < 0) die_write();
      if (!stralloc_append(&buf, &ch)) die_nomem();
    }
    h = cdb_hash(buf.s, klen);
    get(&ch);
    if (ch != '-') die_readformat();
    get(&ch);
    if (ch != '>') die_readformat();
    for (i = 0;i < dlen;++i) {
      get(&ch);
      if (djbio_PUTC(&c.io, ch) < 0) die_write();
    }
    if (cdb_make_addend(&c, klen, dlen, h) < 0) die_write();

    get(&ch);
    if (ch != '\n') die_readformat();
  }

  if (cdb_make_finish(&c) < 0) die_write();
  if (fsync(fd) < 0) die_write();
  if (close(fd) < 0) die_write(); /* NFS silliness */
  if (rename(fntmp, fn) < 0) die_rename();

  _exit(0);
}
